import numpy as np

def cluster_intra_dis(X, cent, label):
    intra_dis = []
    for i in range(len(cent)):
        data_cluster = X[label == i, :]

        cluster_dis = np.linalg.norm(data_cluster - cent[i], axis=0)
        intra_dis.append(np.sum(cluster_dis) / data_cluster.shape[0])

    return intra_dis


def merge_compute(y_pred, mu, scace_emb):

    # Check if len(mu_prepare) == len(np.unique(y_pred))
    idx_cent = []
    for i in range(len(mu)):
        if (y_pred == i).any() == False:
            idx_cent.append(i)
    if len(idx_cent) == 0:
        mu = mu
    else:
        mu = np.delete(mu, idx_cent, 0)
    n_clusters = len(mu)

    # Change label to 0 ~ len(np.unique(label))
    for i in range(len(np.unique(y_pred))):
        if np.unique(y_pred)[i] != i:
            y_pred[y_pred == np.unique(y_pred)[i]] = i


    Centroid = np.array(mu)

    # Compute d_bar
    intra_dis = cluster_intra_dis(scace_emb, Centroid, y_pred)
    d_ave = np.mean(intra_dis)

    sum_1 = 0
    for i in range(0, n_clusters):
        for j in range(i + 1, n_clusters):
            weight = d_ave / ((intra_dis[i] + intra_dis[j]) / 2)
            sum_1 += weight * np.linalg.norm(Centroid[i] - Centroid[j])

    d_bar = sum_1 / (n_clusters * (n_clusters - 1))

    return y_pred, Centroid, d_bar, intra_dis, d_ave



def centroid_merge(X, cent_to_merge, label, min_dis, intra_dis, d_ave):

    pred_f = []

    for t in range(200):

        if t == 0:
            pred_f.append(label)

        Cent_dis = []
        Cent_i = []
        Cent_e = []
        Final_Centroid_merge = cent_to_merge

        n_Clusters = len(Final_Centroid_merge)

        for i in range(n_Clusters):
            for e in range(i + 1, n_Clusters):
                weight = d_ave / ((intra_dis[i] + intra_dis[e]) / 2)
                dis = np.linalg.norm(Final_Centroid_merge[i] - Final_Centroid_merge[e])
                Cent_dis.append(weight * dis)
                Cent_i.append(i)
                Cent_e.append(e)

        for i in range(len(Cent_dis)):
            if Cent_dis[i] < min_dis and Cent_dis[i] == min(Cent_dis):

                Cent_merge = (Final_Centroid_merge[Cent_i[i]] + Final_Centroid_merge[Cent_e[i]]) / 2

                Final_Centroid_merge = np.delete(Final_Centroid_merge, (Cent_i[i], Cent_e[i]), 0)
                Final_Centroid_merge = np.insert(Final_Centroid_merge, Cent_i[i], Cent_merge, 0)
                Final_Centroid_merge = np.insert(Final_Centroid_merge, Cent_e[i], 0, 0)

                Final_Centroid_merge = Final_Centroid_merge[~(Final_Centroid_merge == 0).all(axis=1)]

                cent_to_merge = Final_Centroid_merge

                label = np.array(label)
                label[label == Cent_e[i]] = Cent_i[i]

                for i in range(len(np.unique(label))):
                    if np.unique(label)[i] != i:
                        label[label == np.unique(label)[i]] = i
                    else:
                        continue

            else:
                pass

        n_clusters_t = len(np.unique(label))
        pred_f.append(label)

        intra_dis = cluster_intra_dis(X, cent_to_merge, label)
        d_ave = np.mean(intra_dis)

        sum_1 = 0
        for i in range(n_clusters_t):
            for j in range(i + 1, n_clusters_t):
                weight = d_ave / ((intra_dis[i] + intra_dis[j]) / 2)

                sum_1 += weight * (np.linalg.norm(Final_Centroid_merge[i] - Final_Centroid_merge[j]))

        d_bar = sum_1 / (n_clusters_t * (n_clusters_t - 1))

        min_dis = d_bar

        count = 0
        for i in range(0, n_clusters_t):
            for j in range(i + 1, n_clusters_t):
                weight = d_ave / ((intra_dis[i] + intra_dis[j]) / 2)
                d_inter = weight * np.linalg.norm(Final_Centroid_merge[i] - Final_Centroid_merge[j])

                if d_inter > d_bar:
                    count += 1

        count_true = int((n_clusters_t ** 2 - n_clusters_t) / 2)

        print("-----------------iter: %d-----------------" % int(t + 1))
        print("n_clusters: %d" % n_clusters_t)
        print("count_true: %d" % count_true)
        print("count: %d" % count)

        if count >= count_true:
            print("Reach count!")
            break

        else:
            continue

    return Final_Centroid_merge, label, n_clusters_t, pred_f



def centroid_split(X, X_1, Centroid_split, label):
    """
        Parameters
        ----------
        X
            Embedding after pre-training. Rows are cells and columns are genes.
        X_1
            Embedding + Column vectors of cell types (Label splicing in the last column).
    """

    ### Compute weights
    intra_dis = cluster_intra_dis(X, Centroid_split, label)
    d_ave = np.mean(intra_dis)

    sum_1 = 0
    n_clusters = len(np.unique(label))
    for i in range(0, n_clusters):
        for j in range(i + 1, n_clusters):
            weight = d_ave / ((intra_dis[i] + intra_dis[j]) / 2)
            sum_1 += weight * np.linalg.norm(Centroid_split[i] - Centroid_split[j])

    d_bar = sum_1 / (n_clusters * (n_clusters - 1) / 2)
    min_dis = d_bar / 2

    X_copy = 1 * X_1

    Dis_tol = cluster_intra_dis(X, Centroid_split, label)

    for t in range(200):

        idx_split = []
        for i in range(len(Centroid_split)):
            if Dis_tol[i] > min_dis and Dis_tol[i] == max(Dis_tol):
                idx_split.append(i)
                dis = []
                X_2 = np.delete(X_1[X_1[:, -1] == i], -1, 1)
                for m in range(len(X_2)):
                    dis_append = 0
                    for n in range(len(X_2)):
                        dis_append += np.linalg.norm(X_2[m] - X_2[n]) ** 2
                    dis.append(dis_append)
                idx_1 = np.argmin(dis)
                centriod_1 = X_2[idx_1]

                X_3 = np.delete(X_2, idx_1, 0)
                T_m = []
                for m in range(len(X_3)):
                    T_nm = 0
                    for n in range(len(X_3)):
                        D_n = np.linalg.norm(X_3[n] - centriod_1) ** 2
                        d_nm = np.linalg.norm(X_3[m] - X_3[n]) ** 2
                        T_nm += np.maximum(D_n - d_nm, 0)
                    T_m.append(T_nm)
                idx_2 = np.argmax(T_m)
                centriod_2 = X_3[idx_2]

                centroid = np.concatenate(
                    (centriod_1.reshape(len(centriod_1), 1).T, centriod_2.reshape(len(centriod_2), 1).T), axis=0)
                idx_1 = []
                for j in range(len(X_1[X_1[:, -1] == i])):
                    A = np.delete(X_1[X_1[:, -1] == i][j], -1)
                    distance = []
                    for e in range(2):
                        B = centroid[e]
                        D = np.linalg.norm(A - B)
                        distance.append(D)
                    idx = np.argmin(distance)
                    idx_1.append(idx)

                    if idx == 1:
                        if np.unique(label)[0] == 0:
                            idx_a = np.array(np.where(X_1[:, -1] == i))[0, j]
                            a = X_1[X_1[:, -1] == i]
                            a[j, -1] = len(Centroid_split) + i + 1
                            X_copy[idx_a, :] = a[j, :]

                        else:
                            idx_a = np.array(np.where(X_1[:, -1] == i))[0, j]
                            a = X_1[X_1[:, -1] == i]
                            a[j, -1] = len(Centroid_split) + 2 + i
                            X_copy[idx_a, :] = a[j, :]
                    else:
                        if np.unique(label)[0] == 0:
                            idx_a = np.array(np.where(X_1[:, -1] == i))[0, j]
                            a = X_1[X_1[:, -1] == i]
                            a[j, -1] = len(Centroid_split) + i
                            X_copy[idx_a, :] = a[j, :]

                        else:
                            idx_a = np.array(np.where(X_1[:, -1] == i))[0, j]
                            a = X_1[X_1[:, -1] == i]
                            a[j, -1] = len(Centroid_split) + 1 + i
                            X_copy[idx_a, :] = a[j, :]

                Centroid_split = np.concatenate(
                    (Centroid_split, centriod_1.reshape(1, len(centriod_1)), centriod_2.reshape(1, len(centriod_2))))

            else:
                continue

        if len(idx_split) == 0:
            Centroid_split = Centroid_split
            label = label
        else:
            Centroid_split = np.delete(Centroid_split, idx_split, 0)
            label = X_copy[:, -1]
            label = np.array(label)
            for i in range(len(np.unique(label))):
                if np.unique(label)[i] != i:
                    label[label == np.unique(label)[i]] = i
                else:
                    continue
            label = label.tolist()

        n_clusters = Centroid_split.shape[0]
        X_1 = np.concatenate([np.array(X), np.array(label).reshape(len(label), 1)], axis=1)
        X_copy = 1 * X_1

        ### Compute weights
        intra_dis = cluster_intra_dis(X, Centroid_split, np.array(label))
        d_ave = np.mean(intra_dis)

        sum_1 = 0
        for i in range(0, n_clusters):
            for j in range(i + 1, n_clusters):
                weight = d_ave / ((intra_dis[i] + intra_dis[j]) / 2)
                sum_1 += weight * np.linalg.norm(Centroid_split[i] - Centroid_split[j])

        d_bar = sum_1 / (n_clusters * (n_clusters - 1) / 2)
        min_dis = d_bar / 2

        Dis_tol = cluster_intra_dis(X, Centroid_split, np.array(label))

        count = 0
        for i in range(len(Dis_tol)):
            if Dis_tol[i] < min_dis:
                count += 1

        print("-----------------iter: %d-----------------" % int(t + 1))
        print("n_clusters: %d" % n_clusters)
        print("count_true: %d" % n_clusters)
        print("count: %d" % count)

        if count >= n_clusters:
            print("Reach count!")
            break

        else:
            continue

    return Centroid_split, label




import torch
from torch import nn


class ZINBLoss(nn.Module):
    def __init__(self, ridge_lambda=0):
        super(ZINBLoss, self).__init__()
        self.ridge_lambda = ridge_lambda

    def forward(self, x, mean, disp, pi, scale_factor):
        eps = 1e-10
        mean = mean * scale_factor

        t1 = torch.lgamma(disp + eps) + torch.lgamma(x + 1.0) - torch.lgamma(x + disp + eps)
        t2 = (disp + x) * torch.log(1.0 + (mean / (disp + eps))) + (x * (torch.log(disp + eps) - torch.log(mean + eps)))
        nb_final = t1 + t2

        nb_case = nb_final - torch.log(1.0 - pi + eps)
        zero_nb = torch.pow(disp / (disp + mean + eps), disp)
        zero_case = -torch.log(pi + ((1.0 - pi) * zero_nb) + eps)
        result = torch.where(torch.le(x, 1e-8 * torch.ones_like(x)), zero_case, nb_case)

        if self.ridge_lambda > 0:
            ridge = self.ridge_lambda * torch.square(pi)
            result += ridge

        result = torch.mean(result)

        return result


class ClusterLoss(nn.Module):

    def __init__(self):
        super(ClusterLoss, self).__init__()

    def forward(self, target, pred):
        return torch.mean(torch.sum(target * torch.log(target / (pred + 1e-6)), dim=-1))


class ELOBkldLoss(nn.Module):
    def __init__(self):
        super(ELOBkldLoss, self).__init__()

    def forward(self, mu, logvar):
        result = -((0.5 * logvar) - (torch.exp(logvar) + mu ** 2) / 2. + 0.5)
        result = result.sum(dim=1).mean()

        return result



import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class Mish(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x * torch.tanh(F.softplus(x))


def buildNetwork(layers, activation="relu"):
    net = []
    for i in range(1, len(layers)):
        layer = nn.Linear(layers[i - 1], layers[i])
        nn.init.kaiming_normal_(layer.weight)
        # nn.init.kaiming_normal_(layer.bias)
        nn.init.constant_(layer.bias, 0)
        net.append(layer)

        # net.append(nn.BatchNorm1D(layers[i]))
        if activation == "relu":
            net.append(nn.ReLU())
        elif activation == "sigmoid":
            net.append(nn.Sigmoid())
        elif activation == "mish":
            net.append(Mish())
        elif activation == "tanh":
            net.append(nn.Tanh())
    return nn.Sequential(*net)


class GaussianNoise(nn.Module):
    def __init__(self, sigma=0):
        super(GaussianNoise, self).__init__()
        self.sigma = sigma

    def forward(self, x):
        if self.training:
            x = x + self.sigma * torch.randn_like(x.shape)
        return x


class MeanAct(nn.Module):
    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)


class DispAct(nn.Module):
    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)


class scAce(nn.Module):
    def __init__(self, input_dim, device, z_dim=32,
                 encode_layers=[512, 256],
                 decode_layers=[256, 512],
                 activation='relu'):
        super(scAce, self).__init__()
        self.z_dim = z_dim
        self.activation = activation

        # self.mu = None
        self.pretrain = False
        self.device = device
        self.alpha = 1.
        self.sigma = 0  # 1.

        self.encoder = buildNetwork([input_dim] + encode_layers, activation=activation)
        self.decoder = buildNetwork([z_dim] + decode_layers, activation=activation)

        self.enc_mu = nn.Linear(encode_layers[-1], z_dim)
        self.enc_var = nn.Linear(encode_layers[-1], z_dim)

        self.dec_mean = nn.Sequential(nn.Linear(decode_layers[-1], input_dim), MeanAct())
        self.dec_disp = nn.Sequential(nn.Linear(decode_layers[-1], input_dim), DispAct())
        self.dec_pi = nn.Sequential(nn.Linear(decode_layers[-1], input_dim), nn.Sigmoid())


    def soft_assign(self, z):

        q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - self.mu) ** 2, dim=2))
        q = (q.t() / torch.sum(q, dim=1)).t()
        return q

    def target_distribution(self, q):
        p = q ** 2 / q.sum(0)
        return (p.t() / p.sum(1)).t()

    def reparameterize(self, mu, logvar):

        # obtain standard deviation from log variance
        std = torch.exp(0.5 * logvar)
        # values are sampled from unit normal distribution
        eps = torch.randn(std.shape).to(self.device)
        return mu + eps * std

    def Encoder(self, x):
        # h = self.encoder(x + torch.randn_like(x) * self.sigma)
        h = self.encoder(x)
        z_mu = self.enc_mu(h)
        z_logvar = self.enc_var(h)
        z = self.reparameterize(z_mu, z_logvar)

        if self.pretrain:
            return z_mu, z_logvar, z

        h0 = self.encoder(x)
        z_mu0 = self.enc_mu(h0)
        z_logvar0 = self.enc_var(h0)
        z0 = self.reparameterize(z_mu0, z_logvar0)
        return z_mu, z_logvar, z, z0

    def Decoder(self, z):

        h = self.decoder(z)
        mu = self.dec_mean(h)
        disp = self.dec_disp(h)
        pi = self.dec_pi(h)

        return mu, disp, pi

    def forward(self, x):
        if self.pretrain:
            # Encode
            z_mu, z_logvar, z = self.Encoder(x)

            # Decode
            mu, disp, pi = self.Decoder(z)

            return z_mu, z_logvar, mu, disp, pi

        # else

        # Encode
        z_mu, z_logvar, z, z0 = self.Encoder(x)

        # Decode
        mu, disp, pi = self.Decoder(z)

        # cluster
        q = self.soft_assign(z0)

        return z_mu, z_logvar, mu, disp, pi, q

    def EncodeAll(self, X, batch_size=256):
        all_z_mu = []
        num = X.shape[0]
        num_batch = int(math.ceil(1.0 * X.shape[0] / batch_size))
        for batch_idx in range(num_batch):
            exp = X[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
            exp = torch.tensor(np.float32(exp))
            with torch.no_grad():
                z_mu, _, _, _ = self.Encoder(exp.to(self.device))

            all_z_mu.append(z_mu)

        all_z_mu = torch.cat(all_z_mu, dim=0)
        return all_z_mu


def weight_init(m):
    nn.init.xavier_normal_(m.weight)
    # nn.init.kaiming_normal_(m.bias)
    nn.init.constant_(m.bias, 0)



import os
from setuptools import setup, find_packages

path = os.path.abspath(os.path.dirname(__file__))

try:
    with open(os.path.join(path, 'README.md'), "r", encoding='utf-8') as f:
        long_description = f.read()
except Exception as e:
    long_description = "scAce: an adaptive embedding and clustering method for single-cell gene expression data"

setup(
    name="scace",
    version="0.1.1",
    keywords=["single-cell RNA-sequencing", "clustering", "cluster merging"],
    description="scAce: an adaptive embedding and clustering method for single-cell gene expression data",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT Licence",

    url="https://github.com/sldyns/scAce",
    author="Kun Qian, Xinwei He",
    author_email="kun_qian@foxmail.com",
    maintainer="Kun Qian",
    maintainer_email="kun_qian@foxmail.com",

    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy",
        "scanpy",
        "torch",
        "pandas",
        "tqdm",
        "scipy",
        "sklearn"
        ],
    platforms='any'
)



from .scace_train import run_scace
import importlib

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from pkg_resources import get_distribution
    version = lambda name: get_distribution(name).version

__name__ = "scAce"
try:
    __version__ = version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = '0.1.1'